package org.avaje.metric.filereport;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.Writer;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.avaje.metric.Metric;
import org.avaje.metric.MetricManager;
import org.avaje.metric.MetricVisitor;

/**
 * Writes the collected metrics to a file.
 * <p>
 * Typically you configure the frequency in seconds in which statistics are
 * collected and reported as well as a base directory where the metric files go.
 * By default the base directory will be read from a system property
 * 'metric.directory' and otherwise defaults to the current directory.
 * </p>
 */
public class FileReporter {

  private static final Logger logger = Logger.getLogger(FileReporter.class.getName());

  private static final int DEFAULT_NUM_FILES_TO_KEEP = 20;

  private static final int EIGHT_HOURS = 60 * 60 * 8;

  protected final SimpleDateFormat timeFormat = new SimpleDateFormat("HH:mm:ss");

  protected final Timer timer;

  protected final int freqInSeconds;

  protected final int numberOfFilesToKeep;

  protected final String baseDirectory;

  protected final String baseFileName;

  /**
   * Create using 60 second frequency and defined base directory.
   */
  public FileReporter(String baseDirectory) {
    this(60, baseDirectory);
  }

  /**
   * Create with a freqInSeconds of 60 and using the default base directory.
   */
  public FileReporter() {
    this(60, null);
  }

  /**
   * Create specifying a base directory where the metrics files should go.
   */
  public FileReporter(int freqInSeconds, String baseDirectory) {
    this(freqInSeconds, baseDirectory, null);
  }

  /**
   * Create specifying a write frequency, base directory and base file name.
   */
  public FileReporter(int freqInSeconds, String baseDirectory, String baseFileName) {
    this(freqInSeconds, baseDirectory, baseFileName, -1);
  }

  /**
   * Create specifying a write frequency, base directory and base file name.
   */
  public FileReporter(int freqInSeconds, String baseDirectory, String baseFileName,
      int numberOfFilesToKeep) {

    this.numberOfFilesToKeep = getNumberOfFilesToKeep(numberOfFilesToKeep);
    this.baseDirectory = getBaseDirectory(baseDirectory);
    this.baseFileName = getBaseFileName(baseFileName);

    this.freqInSeconds = freqInSeconds;
    long freqMillis = freqInSeconds * 1000;

    if (isWriteToFile()) {
      // Create the Timer and schedule the WriteTask
      this.timer = new Timer("MetricFileReporter", true);
      this.timer.scheduleAtFixedRate(new WriteTask(), freqMillis, freqMillis);
    } else {
      // Global property to turn of metric writing to file so
      // don't create the Timer and schedule the WriteTask
      this.timer = null;
    }

    deleteOldMetricFilesInBackground();
  }

  /**
   * Cleanup old metric files using a background thread.
   * <p>
   * This is done periodically to try and maintain numberOfFilesToKeep metric
   * files.
   * </p>
   */
  public void deleteOldMetricFilesInBackground() {
    try {
      Thread t = new Thread(new Runnable() {

        @Override
        public void run() {
          deleteOldMetricFiles();
        }
      });

      t.start();
    } catch (Exception e) {
      logger.log(Level.SEVERE, "Error deleting old metrics files in background", e);
    }
  }

  /**
   * Delete old metric files.
   * <p>
   * This is called periodically to cleanup.
   * </p>
   */
  public void deleteOldMetricFiles() {

    try {

      final String minFileName = FileOutput.getFileName(baseFileName, numberOfFilesToKeep);

      File dir = new File(baseDirectory);
      String[] delFileNames = dir.list(new FilenameFilter() {

        @Override
        public boolean accept(File dir, String name) {
          if (name.startsWith(baseFileName) && name.compareTo(minFileName) < 0) {
            return true;
          }
          return false;
        }
      });

      for (int i = 0; i < delFileNames.length; i++) {
        File f = new File(dir, delFileNames[i]);
        if (f.exists()) {
          f.delete();
        }
      }

    } catch (Exception e) {
      logger.log(Level.SEVERE, "Error trying to cleanup old metric files", e);
    }
  }

  /**
   * Return true if the metric.writeToFile system property allows writing.
   */
  protected boolean isWriteToFile() {
    String value = System.getProperty("metric.writeToFile");
    if (value != null && value.trim().toLowerCase().equals("false")) {
      // explicitly disabled metric writing to file
      return false;
    }
    return true;
  }

  protected static int getNumberOfFilesToKeep(int value) {
    if (value < 1) {
      String sysVal = System.getProperty("metric.numberOfFilesToKeep");
      if (sysVal != null) {
        try {
          return Integer.parseInt(sysVal);
        } catch (NumberFormatException e) {
          return DEFAULT_NUM_FILES_TO_KEEP;
        }
      }
    }

    return DEFAULT_NUM_FILES_TO_KEEP;
  }

  /**
   * Return the base directory to put metrics files in - defaults to current
   * working directory.
   */
  protected String getBaseDirectory(String baseDirectory) {
    if (baseDirectory == null) {
      baseDirectory = System.getProperty("metric.directory");
    }
    if (baseDirectory == null) {
      baseDirectory = ".";
    }
    return baseDirectory;
  }

  /**
   * Return the base file name - defaults to "metric".
   */
  protected String getBaseFileName(String baseFileName) {
    if (baseFileName == null) {
      baseFileName = System.getProperty("metric.file");
    }
    if (baseFileName == null) {
      baseFileName = "metric";
    }
    return baseFileName;
  }

  protected class WriteTask extends TimerTask {

    int cleanupCounter;

    public void run() {
      try {
        cleanupCounter++;
        writeAll();

        if (cleanupCounter * freqInSeconds > EIGHT_HOURS) {
          // cleanup old metric files about every 8 hours
          cleanupCounter = 0;
          deleteOldMetricFilesInBackground();
        }

      } catch (IOException e) {
        logger.log(Level.SEVERE, "", e);
      }
    }
  }

  /**
   * Write all the metrics appending to the appropriate file.
   */
  protected void writeAll() throws IOException {

    FileOutput fo = new FileOutput(baseDirectory, baseFileName);
    try {
      Writer writer = fo.getWriter();

      TextFileWriteVisitor visitor = new TextFileWriteVisitor(writer, freqInSeconds);
      visitor.onStartVisiting();

      Collection<Metric> visitOnlyMetrics = MetricManager.getJvmMetrics();
      visitSorted(visitOnlyMetrics, visitor);

      Collection<Metric> metrics = MetricManager.getMetrics();
      visitSorted(metrics, visitor);

    } finally {
      fo.close();
    }
  }

  /**
   * Visit the metrics sorted by name.
   */
  private void visitSorted(Collection<Metric> allMetrics, MetricVisitor visitor) {
    ArrayList<Metric> ar = new ArrayList<Metric>(allMetrics);
    Collections.sort(ar, new NameComp());

    for (Metric metric : ar) {
      metric.visit(visitor);
    }
  }

  /**
   * Compare Metrics by name.
   */
  private static class NameComp implements Comparator<Metric> {

    @Override
    public int compare(Metric o1, Metric o2) {
      return o1.getName().compareTo(o2.getName());
    }

  }

}
